use std::collections::HashMap;
use std::iter::zip;

use hercules_ir::callgraph::*;
use hercules_ir::def_use::*;
use hercules_ir::ir::*;

use crate::*;

/*
 * Top level function to run inlining. Currently, inlines every function call,
 * since mutual recursion is not valid in Hercules IR.
 */
pub fn inline(editors: &mut [FunctionEditor], callgraph: &CallGraph) {
    // Step 1: run topological sort on the call graph to inline the "deepest"
    // function first.
    let topo = callgraph.topo();

    // Step 2: make sure each function has a single return node. If an edit
    // failed to make a function have a single return node, then we can't inline
    // calls of it.
    let single_return_nodes: Vec<_> = editors
        .iter_mut()
        .map(|editor| collapse_returns(editor))
        .collect();

    // Step 3: verify that each possible dynamic constant parameter index has a
    // single unique dynamic constant ID. If this isn't true, dynamic constant
    // substitution won't work, and this should be true anyway!
    let mut found_idxs = HashMap::new();
    for id in editors[0].dynamic_constant_ids() {
        let dc = editors[0].get_dynamic_constant(id);
        if let DynamicConstant::Parameter(idx) = *dc {
            assert!(!found_idxs.contains_key(&idx));
            found_idxs.insert(idx, id);
        }
    }
    let mut dc_param_idx_to_dc_id = vec![];
    for idx in 0..found_idxs.len() {
        dc_param_idx_to_dc_id.push(found_idxs[&idx]);
    }

    // Step 4: run inlining on each function individually. Iterate the functions
    // in topological order.
    for to_inline_id in topo {
        // Since Rust cannot analyze the accesses into an array of mutable
        // references, we need to do some weirdness here to simultaneously get:
        // 1. A mutable reference to the function we're modifying.
        // 2. Shared references to all of the functions called by that function.
        let callees = callgraph.get_callees(to_inline_id);
        let editor_refs = get_mut_and_immuts(editors, to_inline_id, callees);
        inline_func(
            editor_refs.0,
            editor_refs.1,
            &single_return_nodes,
            &dc_param_idx_to_dc_id,
        );
    }
}

/*
 * Helper function to get from an array of mutable references:
 * 1. A single mutable reference.
 * 2. Several shared references.
 * Where none of the references alias. We need to use this both for function
 * editors and schedules.
 */
fn get_mut_and_immuts<'a, T, I: ID>(
    mut_refs: &'a mut [T],
    mut_id: I,
    shared_id: &[I],
) -> (&'a mut T, HashMap<I, &'a T>) {
    let mut all_id = Vec::from(shared_id);
    all_id.sort_unstable();
    all_id.insert(all_id.binary_search(&mut_id).unwrap_err(), mut_id);
    let mut mut_ref = None;
    let mut shared_refs = HashMap::new();
    let mut cursor = 0;
    let mut slice = &mut *mut_refs;
    for id in all_id {
        let (left, right) = slice.split_at_mut(id.idx() - cursor);
        cursor += left.len() + 1;
        let (left, right) = right.split_at_mut(1);
        let item = &mut left[0];
        if id == mut_id {
            assert!(mut_ref.is_none());
            mut_ref = Some(item);
        } else {
            shared_refs.insert(id, &*item);
        }
        slice = right;
    }

    (mut_ref.unwrap(), shared_refs)
}

/*
 * Run inlining on a single function. Pass a mutable reference to the function
 * to modify and shared references for all called functions.
 */
fn inline_func(
    editor: &mut FunctionEditor,
    called: HashMap<FunctionID, &FunctionEditor>,
    single_return_nodes: &Vec<Option<NodeID>>,
    dc_param_idx_to_dc_id: &Vec<DynamicConstantID>,
) {
    let first_num_nodes = editor.func().nodes.len();
    for id in (0..first_num_nodes).map(NodeID::new) {
        // Break down the call node.
        let Node::Call {
            control,
            function,
            ref dynamic_constants,
            ref args,
        } = editor.func().nodes[id.idx()]
        else {
            continue;
        };

        // Assemble all the info we'll need to do the edit.
        let dcs_a = &dc_param_idx_to_dc_id[..dynamic_constants.len()];
        let dcs_b = dynamic_constants.clone();
        let args = args.clone();
        let old_num_nodes = editor.func().nodes.len();
        let old_id_to_new_id = |old_id: NodeID| NodeID::new(old_id.idx() + old_num_nodes);
        let call_pred = get_uses(&editor.func().nodes[control.idx()]);
        assert_eq!(call_pred.as_ref().len(), 1);
        let call_pred = call_pred.as_ref()[0];
        let called_func = called[&function].func();
        // We can't inline calls to functions with multiple returns.
        let Some(called_return) = single_return_nodes[function.idx()] else {
            continue;
        };
        let called_return_uses = get_uses(&called_func.nodes[called_return.idx()]);
        let called_return_pred = called_return_uses.as_ref()[0];
        let called_return_data = called_return_uses.as_ref()[1];

        // Perform the actual edit.
        editor.edit(|mut edit| {
            // Insert the nodes from the called function. There are a few
            // special cases:
            // - Start: don't add start nodes - later, we'll replace_all_uses on
            //   the start node with the one predecessor of the call's region
            //   node.
            // - Parameter: don't add parameter nodes - later, we'll
            //   replace_all_uses on the parameter nodes with the arguments to
            //   the call node.
            // - Return: don't add return nodes - later, we'll replace_all_uses
            //   on the call's region node with the predecessor to the return
            //   node.
            for (idx, node) in called_func.nodes.iter().enumerate() {
                if node.is_start() || node.is_parameter() || node.is_return() {
                    // We still need to add some node to make sure the IDs line
                    // up. Just add a gravestone.
                    edit.add_node(Node::Start);
                    continue;
                }
                // Get the node from the callee function and replace all the
                // uses with the to-be IDs in the caller function.
                let mut node = node.clone();
                if node.is_fork()
                    || node.is_constant()
                    || node.is_dynamic_constant()
                    || node.is_call()
                {
                    // We have to perform the subsitution in two steps. First,
                    // we map every dynamic constant A to a non-sense dynamic
                    // constant ID. Second, we map each non-sense dynamic
                    // constant ID to the appropriate dynamic constant B. Why
                    // not just do this in one step from A to B? We update
                    // dynamic constants one at a time, so imagine the following
                    // A -> B mappings:
                    // ID 0 -> ID 1
                    // ID 1 -> ID 0
                    // First, we apply the first mapping. This changes all
                    // references to dynamic constant 0 to dynamic constant 1.
                    // Then, we apply the second mapping. This updates all
                    // already present references to dynamic constant 1, as well
                    // as the new references we just made in the first step. We
                    // actually want to institute all the updates
                    // *simultaneously*, hence the two step maneuver.
                    let first_dc = edit.num_dynamic_constants() + 10000;
                    for (dc_a, dc_n) in zip(dcs_a, first_dc..) {
                        substitute_dynamic_constants_in_node(
                            *dc_a,
                            DynamicConstantID::new(dc_n),
                            &mut node,
                            &mut edit,
                        );
                    }
                    for (dc_n, dc_b) in zip(first_dc.., dcs_b.iter()) {
                        substitute_dynamic_constants_in_node(
                            DynamicConstantID::new(dc_n),
                            *dc_b,
                            &mut node,
                            &mut edit,
                        );
                    }
                }
                let mut uses = get_uses_mut(&mut node);
                for u in uses.as_mut() {
                    **u = old_id_to_new_id(**u);
                }
                // Add the node and check that the IDs line up.
                let add_id = edit.add_node(node);
                assert_eq!(add_id, old_id_to_new_id(NodeID::new(idx)));
                // Copy the schedule from the callee.
                let callee_schedule = &called_func.schedules[idx];
                for schedule in callee_schedule {
                    edit = edit.add_schedule(add_id, schedule.clone())?;
                }
                // Copy the labels from the callee.
                let callee_labels = &called_func.labels[idx];
                for label in callee_labels {
                    edit = edit.add_label(add_id, *label)?;
                }
            }

            // Stitch the control use of the inlined start node with the
            // predecessor control node of the call's region.
            let start_node = &called_func.nodes[0];
            assert!(start_node.is_start());
            let start_id = old_id_to_new_id(NodeID::new(0));
            edit = edit.replace_all_uses(start_id, call_pred)?;

            // Stich the control use of the original call node's region with
            // the predecessor control of the inlined function's return.
            edit = edit.replace_all_uses(
                control,
                if called_return_pred == NodeID::new(0) {
                    call_pred
                } else {
                    old_id_to_new_id(called_return_pred)
                },
            )?;

            // Stitch uses of parameter nodes in the inlined function to the IDs
            // of arguments provided to the call node.
            for (node_idx, node) in called_func.nodes.iter().enumerate() {
                if let Node::Parameter { index } = node {
                    let param_id = old_id_to_new_id(NodeID::new(node_idx));
                    edit = edit.replace_all_uses(param_id, args[*index])?;
                }
            }

            // Finally, delete the call node.
            edit = edit.replace_all_uses(id, old_id_to_new_id(called_return_data))?;
            edit = edit.delete_node(control)?;
            edit = edit.delete_node(id)?;

            Ok(edit)
        });
    }
}
